import datetime
import logging
import time

import attr
import pkg_resources
import requests
import six
import udatetime
import yaml


logger = logging.getLogger(__name__)


@attr.s(slots=True)
class AuthInfo(object):
    # Rationale: Storing and maintining expiration time is easier
    # (IMO) than relying on 401s and tracking failure paths.
    access_token = attr.ib(repr=False)
    expires_at = attr.ib()
    scope = attr.ib()
    token_type = attr.ib()

    @property
    def rttl(self):
        """remaining time to live"""
        return self.expires_at - time.time()


@six.python_2_unicode_compatible
class HelixClient(object):
    """Provides a Pythonic interface to the Helix identity and genomics
    REST-based v0 API.

    Features:
    * transparent, on-demand reauthentication for expired tokens
    * unification across Helix scopes
    * streaming (generator) interface from paged replies

    For the Helix demo credentials:

    >>>  hc = HelixClient(
    ...       base_urls={
    ...           'genomics': 'https://genomics.staging.helix.com/v0/',
    ...           'identity': 'https://api.staging.helix.com/v0/'},
    ...       credentials={
    ...           'genomics': {
    ...               'client_id': 'CL-AQ3EJLBNNDE7LJZYVGUOMM4WA5HBOHEL',
    ...               'client_secret': 'CS-Y5MFWTTZU67LOIEBOWFT2LE7DEI3OPP7'},
    ...           'identity': {
    ...               'client_id': 'CL-7ICCDTTMGWK7VHTJIHCPWZZ7EEZBID7Z',
    ...               'client_secret': 'CS-B7XBHYQLWURKBV3ZXVBGK7SXIZ7JGKTG'}})

    """


    def __init__(self, env, credentials):
        """Params:

        :param str env: helix environment (`staging` or `production`)
        :param dict credentials: dict of {service: cred_dict}, where
            cred_dict looks like {`client_id`: str, `client_secret`:
            str}

        HelixClient provides a simple, unified interface for Helix API
        "services" (identity and genetics for now) for a single
        application.

        """

        self._base_urls = yaml.load(pkg_resources.resource_stream("helix", "_data/base-urls.yml"))[env]
        self._credentials = credentials
        self._authinfos = {}
        self._logger = logging.getLogger(__name__)


    def __str__(self):
        """also provides __unicode__() implicitly via
        @six.python_2_unicode_compatible *class* decorator

        """
        return "{}(identity url = {})".format(type(self).__name__,
                                              self._base_urls["identity"])



    def get_events(self, after, event_type=None, until=None):

        """provide a streaming interface on /events

        :param uint after: return events strictly after this sequence
        number
        :param str event_type: only return events of this type
        :param str|timestamp until: return events up to and including this
        timestamp (e.g., `2016-12-01T00:00:00Z` or 1480550400.0)

        :returns: an iterator over a stream of Helix events, each
        returned as a dictionaries. HelixClient automatically pages
        results and refreshes tokens as needed.

        :raises StopIteration: when event stream is exhausted

        Usage note: When reading a stream up to the `until` date, the
        last returned `event["sequenceNum"]` value should be used for
        a subsequent call to `get_events()`.

        """

        limit = 1000
        last_snum = after

        if until and isinstance(until, str):
            until = udatetime.from_string(until)
            print("until = " + str(until))

        while True:
            params = {"after": last_snum, "limit": limit}
            if event_type:
                params["eventType"] = event_type
            events = self._get(rel_url="events", scope="identity",
                               params=params)["events"]
            for e in events:
                if until and udatetime.from_string(e["timestamp"]) > until:
                    raise StopIteration()
                assert e["sequenceNum"] > last_snum, "events are expected to be returned in order"
                last_snum = e["sequenceNum"]
                yield e
            if len(events) < limit:
                raise StopIteration()


    def get_customer(self, pacid):
        """returns a dictionary of customer information for the specified pacid

        :param str pacid: the Helix partner-application-customer ID
        """
        return self._get(rel_url="customers/" + pacid, scope="identity")


    def get_sample_status(self, pacid):
        """returns a list of sample status events for the specified pacid

        :param str pacid: the Helix partner-application-customer ID
        """
        return self._get(rel_url="customers/" + pacid + "/samplestatus",
                         scope="identity")["sampleStatus"]


    def get_variants(self, pacid):
        """returns a list of variants for the specified pacid

        :param str pacid: the Helix partner-application-customer ID
        """
        return self._post(rel_url="variants/search",
                          scope="genomics",
                          data='{{ "callSetIds": ["{pacid}"] }}'.format(pacid=pacid))



    # #####################################################################
    # INTERNAL METHODS

    def _get(self, rel_url, scope, params=None, data=None):
        return self._do(requests.get, rel_url, scope, params, data)


    def _post(self, rel_url, scope, params=None, data=None):
        return self._do(requests.post, rel_url, scope, params, data)


    def _do(self, m, rel_url, scope, params=None, data=None):
        url = self._base_urls[scope] + rel_url
        token = self._fetch_token_for_scope(scope)
        headers = {"Authorization": "Bearer " + token}
        logger.debug("equiv: curl -X {m} -H 'Authorization: {headers[Authorization]}' -d '{data}' {url}".format(
            m=m.__name__.upper(), headers=headers, data=data, url=url))
        r = m(url, headers=headers, params=params, data=data)
        r.raise_for_status()
        return r.json()


    def _fetch_token_for_scope(self, scope, min_rttl=60):
        """Returns a valid access_token for the given `scope` from a token
        cache.

        If the cached token doesn't exist or is within `min_rttl`
        seconds of expiration ("remaining time to live"), a new token
        is obtained and cached.

        """

        def _get_token(base_url, scope, id, secret):
            url = base_url + "oauth/access_token"
            auth = requests.auth.HTTPBasicAuth(id, secret)
            d = {"grant_type": "client_credentials", "scope": scope}
            r = requests.post(url, auth=auth, data=d)
            r.raise_for_status()
            j = r.json()
            return AuthInfo(
                access_token = j["access_token"],
                expires_at = j["expires_in"] + time.time(),
                scope = j["scope"],
                token_type = j["token_type"],
            )

        if scope in self._authinfos and self._authinfos[scope].rttl < min_rttl:
            self._logger.info("Expiring {ais} (rttl: {ais.rttl})".format(ais=self._authinfos[scope]))
            self._authinfos.pop(scope)

        if scope not in self._authinfos:
            self._logger.info("fetching token for " + scope)
            self._authinfos[scope] = _get_token(self._base_urls["identity"], scope,
                                                self._credentials[scope]["client_id"],
                                                self._credentials[scope]["client_secret"])

        return self._authinfos[scope].access_token



if __name__ == "__main__":      # pragma: nocover
    logging.basicConfig(level=logging.INFO)
    logger = logging.getLogger(__name__)

    env = "staging"
    credentials = yaml.load(pkg_resources.resource_stream("helix", "_data/demo-credentials.yml"))

    hc = HelixClient(env=env, credentials=credentials[env])

    events_i = hc.get_events(0)
    events = list(events_i)
    pacids = sorted(set(e["pacId"] for e in events))
    eventtypes = sorted(set(e["eventType"] for e in events))
    logger.info("Fetched {} events with {} pacIds and {} eventTypes ".format(
        len(events), len(pacids), len(eventtypes)))

    # only pacid with customer and sample status:
    pacid = "PC-VCIOSK47ERXXIBPO5ZBOMW2HSNCQETCY"

    customer = hc.get_customer(pacid)
    logger.info("Customer = " + str(customer))

    sample_status = hc.get_sample_status(pacid)
    logger.info("{} sample status updates for {}".format(len(sample_status), pacid))

    variants = hc.get_variants(pacid)
    assert len(set(v["variantSetId"] for v in variants)) == 1, "expected variants for exactly 1 variantSetId"
    logger.info("{} sample variants for {}".format(len(variants), pacid))

    vset_ids = sorted(set(v["variantSetId"] for v in variants))
